<#
 Copyright (c) 2009-2015 Oliver Mueller [http://oliver-mueller.com]
 
 This file is part of WOPR.
 
 $Id: Create-Server-Certificate.ps1 82 2015-11-10 04:39:37Z oliver $
 
 WOPR is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
#>

<#
.SYNOPSIS
Creates key and X.509 certificate for WOPR server on Windows.

.DESCRIPTION
This script uses OpenSSL to create a private key and a self-signed
X.509 certificate for the WOPR server.

This script searches for the openssl program in the PATH. Alternatively you
can provide the location of the openssl program by option -openssl.

It is assumed that the key file will be called server.key, the certificate
signing request server.csr, and the certificate file server.crt. If you want
you can override these settings by the optional parameters -SSLKey, -SSLCsr,
and -SSLCrt.

Create-Server-Certificate uses a default SSL subject unless you provide your
own value via -SSLSubject.

.EXAMPLE

PS C:\wopr-3.00\vs2013> .\create-server-certificate.ps1

This command creates the key and server certificate with using openssl.exe
from PATH. It creates server.key, server.csr, and server.crt in the current
working directory.

.EXAMPLE

PS C:\wopr-3.00\vs2013> .\create-server-certificate.ps1 -openssl=C:\OpenSSL\bin\openssl.exe

This command creates the key and server certificate with using the specified
OpenSSL program. It creates server.key, server.csr, and server.crt in the current
working directory.

.EXAMPLE

PS C:\wopr-3.00\vs2013> .\create-server-certificate.ps1 -SSLKey=foo.key -SSLCsr=foo.csr -SSLCrt=foo.crt

This command creates the key and server certificate with using the specified
OpenSSL program. It creates foo.key, foo.csr, and foo.crt in the current
working directory.
#>

# Parameters
param(
    [Parameter(Mandatory=$false, ValueFromPipeline=$false)]
    [string]$SSLKey="server.key",

    [Parameter(Mandatory=$false, ValueFromPipeline=$false)]
    [string]$SSLCsr="server.csr",

    [Parameter(Mandatory=$false, ValueFromPipeline=$false)]
    [string]$SSLCrt="server.crt",

    [Parameter(Mandatory=$false, ValueFromPipeline=$false)]
    [string]$SSLSubject = "/C=US/ST=Colorado/L=Cheyenne Mountain/O=USAF/OU=NORAD/CN=wopr.norad.mil",
    
    [ValidateScript({Test-Path -PathType Leaf $_})]
    [Parameter(Mandatory=$false, ValueFromPipeline=$false)]
    [string]$openssl
)

# Process parameters
if ("X$openssl" -eq 'X') {
    $OpenSSLCmd = 'openssl'
} else {
    $OpenSSLCmd = "$openssl"
}

Write-Output 'Generating private key'
& "$OpenSSLCmd" req -new -newkey rsa:2048 -out $SSLCsr -nodes -keyout $SSLKey -days 3650 -subj $SSLSubject

if ($LastExitCode -eq 0) {
    Write-Output 'Generating certificate'
    & "$OpenSSLCmd" x509 -req -in $SSLCsr -out $SSLCrt -signkey $SSLKey -days 3650
}
