/*
* Copyright (c) 2009-2015 Oliver Mueller [http://oliver-mueller.com]
*
* This file is part of WOPR.
*
* $Id: sterminal.cpp 68 2015-07-22 15:15:56Z oliver $
*
* WOPR is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <boost/algorithm/string.hpp>
#include <boost/bind.hpp>
#include <boost/asio.hpp>
#include <boost/asio/ssl.hpp>
#include <boost/program_options.hpp>
#include "config.h"

class WoprPortableTerminal
{
public:

	WoprPortableTerminal(boost::asio::io_service& io_service, boost::asio::ssl::context& context,
		boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
		: socket_(io_service, context)
	{
		boost::asio::ip::tcp::endpoint endpoint = *endpoint_iterator;
		disconnect_ = false;
		socket_.lowest_layer().async_connect(endpoint,
			boost::bind(&WoprPortableTerminal::handle_connect, this,
			boost::asio::placeholders::error, ++endpoint_iterator));
	}

	void handle_connect(const boost::system::error_code& error,
		boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
	{
		if (!error)
		{
			socket_.async_handshake(boost::asio::ssl::stream_base::client,
				boost::bind(&WoprPortableTerminal::handle_handshake, this,
				boost::asio::placeholders::error));
		}
		else if (endpoint_iterator != boost::asio::ip::tcp::resolver::iterator())
		{
			socket_.lowest_layer().close();
			boost::asio::ip::tcp::endpoint endpoint = *endpoint_iterator;
			socket_.lowest_layer().async_connect(endpoint,
				boost::bind(&WoprPortableTerminal::handle_connect, this,
				boost::asio::placeholders::error, ++endpoint_iterator));
		}
		else
		{
			std::cout << "Connect failed: " << error.message() << std::endl;
			std::cout << error.category().name() << ':' << error.value() << std::endl;
		}
	}

	void handle_handshake(const boost::system::error_code& error)
	{
		if (!error)
		{
			socket_.async_read_some(boost::asio::buffer(reply_, max_length),
				boost::bind(&WoprPortableTerminal::handle_read, this,
				boost::asio::placeholders::error,
				boost::asio::placeholders::bytes_transferred));
		}
		else
		{
			std::cout << "Handshake failed: " << error.message() << std::endl;
			std::cout << error.category().name() << ':' << error.value() << std::endl;
		}
	}

	void handle_write(const boost::system::error_code& error,
		size_t bytes_transferred)
	{
		if (!error)
		{
			socket_.async_read_some(boost::asio::buffer(reply_, max_length),
				boost::bind(&WoprPortableTerminal::handle_read, this,
				boost::asio::placeholders::error,
				boost::asio::placeholders::bytes_transferred));
		}
		else
		{
			/*
			* WOPR server on Unix-like systems:
			* ---------------------------------
			* Category == ssl && error == 335544539 => connection closed.
			* Category == misc && error == 2 => EOF => connection closed.
			*
			* WOPR server on Windows:
			* -----------------------
			* Category == system && error == 104 => connection reset by peer.
			* Category == system && error == 10054 => connection closed.
			*/
			if ((error.category() == boost::asio::error::get_ssl_category() && error.value() == 335544539) ||
				(error.category() == boost::asio::error::get_misc_category() && error.value() == 2) ||
				(error.category() == boost::asio::error::get_system_category() && error.value() == 104) ||
				(error.category() == boost::asio::error::get_system_category() && error.value() == 10054))
			{
				std::cout << "*** LOST CONNECTION TO REMOTE HOST ***" << std::endl << std::endl;
			}
			else
			{
				std::cout << "Write failed: " << error.message() << std::endl;
				std::cout << error.category().name() << ':' << error.value() << std::endl;
				std::cout << "SSL: " << boost::asio::error::get_ssl_category().name() << std::endl;
				std::string err = error.message();
				if (error.category() == boost::asio::error::get_ssl_category())
				{
					err = std::string(" (")
						+ boost::lexical_cast<std::string>(ERR_GET_LIB(error.value())) + ","
						+ boost::lexical_cast<std::string>(ERR_GET_FUNC(error.value())) + ","
						+ boost::lexical_cast<std::string>(ERR_GET_REASON(error.value())) + ") "
						;
					char buf[128];
					::ERR_error_string_n(error.value(), buf, sizeof(buf));
					err += buf;
				}
				std::cout << err << std::endl;
			}
		}
	}

	void handle_read(const boost::system::error_code& error,
		size_t bytes_transferred)
	{
		if (!error)
		{
			// Write server's reply to stdout.
			std::cout.write(reply_, bytes_transferred);

			// Read new data from stdin only if command "disconnect"
			// was not entered before.
			if (!disconnect_)
			{
				// Read command from stdin.
				std::cin.getline(request_, max_length);
				size_t request_length = strlen(request_);

				// Check for know commands.
				std::string request(request_);
				boost::trim(request);
				boost::to_upper(request);
				if (request.compare("DISCONNECT") == 0)
				{
					disconnect_ = true;
				}

				// Send command to server.
				boost::asio::async_write(socket_,
					boost::asio::buffer(request_, request_length),
					boost::bind(&WoprPortableTerminal::handle_write, this,
					boost::asio::placeholders::error,
					boost::asio::placeholders::bytes_transferred));
			}
			else
			{
				// User has entered "disconnect" already so let's
				// read data until server closes connection.
				socket_.async_read_some(boost::asio::buffer(reply_, max_length),
					boost::bind(&WoprPortableTerminal::handle_read, this,
					boost::asio::placeholders::error,
					boost::asio::placeholders::bytes_transferred));
			}
		}
		else
		{
			/*
			 * WOPR server on Unix-like systems:
			 * ---------------------------------
			 * Category == ssl && error == 335544539 => connection closed.
			 * Category == misc && error == 2 => EOF => connection closed.
			 *
			 * WOPR server on Windows:
			 * -----------------------
			 * Category == system && error == 104 => connection reset by peer.
			 * Category == system && error == 10054 => connection closed.
			 */
			if ((error.category() == boost::asio::error::get_ssl_category() && error.value() == 335544539) ||
				(error.category() == boost::asio::error::get_misc_category() && error.value() == 2) ||
				(error.category() == boost::asio::error::get_system_category() && error.value() == 104) ||
				(error.category() == boost::asio::error::get_system_category() && error.value() == 10054))
			{
				if (disconnect_)
				{
					std::cout << "*** DISCONNECTED ***" << std::endl << std::endl;
				}
				else
				{
					std::cout << "*** CONNECTION CLOSED BY REMOTE HOST ***" << std::endl << std::endl;
				}
			}
			else
			{
				std::cout << "Read failed: " << error.message() << std::endl;
				std::cout << error.category().name() << ':' << error.value() << std::endl;
				std::cout << "SSL: " << boost::asio::error::get_ssl_category().name() << std::endl;
				std::string err = error.message();
				if (error.category() == boost::asio::error::get_ssl_category())
				{
					err = std::string(" (")
						+ boost::lexical_cast<std::string>(ERR_GET_LIB(error.value())) + ","
						+ boost::lexical_cast<std::string>(ERR_GET_FUNC(error.value())) + ","
						+ boost::lexical_cast<std::string>(ERR_GET_REASON(error.value())) + ") "
						;
					char buf[128];
					::ERR_error_string_n(error.value(), buf, sizeof(buf));
					err += buf;
				}
				std::cout << err << std::endl;
			}
		}
	}

	enum
	{
		max_length = 1024
	};

private:
	boost::asio::ssl::stream<boost::asio::ip::tcp::socket> socket_;
	char request_[max_length];
	char reply_[max_length];
	bool disconnect_;
};

int main(int argc, char* argv[])
{
	boost::program_options::options_description desc(
		"Usage: sterminal [--host <host> | -h <host>] [--port <port> | -p <port>]\\\n"
		"                 [--cert <file> | -c <file>]\n"
		"       sterminal --help\n"
		"       sterminal --version\n"
		"where"
		);
	desc.add_options()
		("help", "Print this help message.")
		("version", "Print version information and exit.")
		("host,h", boost::program_options::value<std::string>(), "Host to connect to. (Default: localhost)")
		("port,p", boost::program_options::value<std::string>(), "Port to connect to. (Default: 7777)")
		("cert,c", boost::program_options::value<std::string>(), "Certificate file for SSL. (Default: server.crt)")
		;
	boost::program_options::variables_map vm;

	try
	{
		boost::program_options::store(boost::program_options::parse_command_line(argc, argv, desc), vm);
		boost::program_options::notify(vm);

		if (vm.count("help"))
		{
			std::cout << desc << std::endl;
			return 1;
		}

		if (vm.count("version"))
		{
			std::cout << "WOPR Terminal (portable)" << std::endl
				<< "Version " << VERSION << std::endl;
			return 0;
		}

		boost::asio::io_service io_service;
		std::string cert_file("server.crt");
		std::string host("localhost");
		std::string port("7777");

		if (vm.count("cert"))
		{
			cert_file = vm["cert"].as<std::string>();
		}

		if (vm.count("host"))
		{
			host = vm["host"].as<std::string>();
		}

		if (vm.count("port"))
		{
			port = vm["port"].as<std::string>();
		}

		boost::asio::ip::tcp::resolver resolver(io_service);
		boost::asio::ip::tcp::resolver::query query(host, port);
		boost::asio::ip::tcp::resolver::iterator iterator = resolver.resolve(query);

		boost::asio::ssl::context ctx(io_service, boost::asio::ssl::context::sslv23);
		ctx.set_verify_mode(boost::asio::ssl::context::verify_peer);
		ctx.load_verify_file(cert_file);

		WoprPortableTerminal terminal(io_service, ctx, iterator);

		io_service.run();
	}
	catch (std::exception& e)
	{
		std::cerr << "Exception: " << e.what() << std::endl;
	}

	return 0;
}
