/*
 * Copyright (c) 2009-2015 Oliver Mueller [http://oliver-mueller.com]
 * 
 * This file is part of WOPR.
 * 
 * $Id: terminal.cpp 56 2015-07-13 02:51:30Z oliver $
 * 
 * WOPR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <boost/algorithm/string.hpp>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/asio.hpp>
#include <boost/asio/ssl.hpp>
#include <boost/program_options.hpp>
#include "config.h"

#ifdef sun
#include <cstdio>
#endif

class WoprTerminal
{
public:

    WoprTerminal(boost::asio::io_service& io_service, boost::asio::ssl::context& context,
                 boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
    : io_service_(io_service),
    socket_(io_service, context),
#ifdef sun
    input_(io_service, (isatty(fileno(stdin)) ? ::open("/dev/tty", O_RDONLY | O_NONBLOCK) : ::dup(STDIN_FILENO))),
#else
    input_(io_service, ::dup(STDIN_FILENO)),
#endif
    input_buffer_(WoprTerminal::max_length)
    {
        disconnect_ = false;
        boost::asio::ip::tcp::endpoint endpoint = *endpoint_iterator;
        socket_.lowest_layer().async_connect(endpoint,
                                             boost::bind(&WoprTerminal::handle_connect, this,
                                                         boost::asio::placeholders::error, ++endpoint_iterator));
    }

    enum
    {
        max_length = 1024
    };

protected:

    void handle_connect(const boost::system::error_code& error,
                        boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
    {
        if (!error)
        {
            socket_.async_handshake(boost::asio::ssl::stream_base::client,
                                    boost::bind(&WoprTerminal::handle_handshake, this,
                                                boost::asio::placeholders::error));
        }
        else if (endpoint_iterator != boost::asio::ip::tcp::resolver::iterator())
        {
            socket_.lowest_layer().close();
            boost::asio::ip::tcp::endpoint endpoint = *endpoint_iterator;
            socket_.lowest_layer().async_connect(endpoint,
                                                 boost::bind(&WoprTerminal::handle_connect, this,
                                                             boost::asio::placeholders::error, ++endpoint_iterator));
        }
        else
        {
            std::cout << "Connect failed: " << error.message() << std::endl;
            std::cout << error.category().name() << ':' << error.value() << std::endl;
        }
    }

    void handle_handshake(const boost::system::error_code& error)
    {
        if (!error)
        {
            boost::asio::async_read_until(input_, input_buffer_, '\n',
                                          boost::bind(&WoprTerminal::handle_read_input, this,
                                                      boost::asio::placeholders::error,
                                                      boost::asio::placeholders::bytes_transferred));
            socket_.async_read_some(boost::asio::buffer(reply_, max_length),
                                    boost::bind(&WoprTerminal::listen_on_server, this,
                                                boost::asio::placeholders::error,
                                                boost::asio::placeholders::bytes_transferred));
        }
        else
        {
            std::cout << "Handshake failed: " << error.message() << std::endl;
            std::cout << error.category().name() << ':' << error.value() << std::endl;
        }
    }

    void listen_on_server(const boost::system::error_code& error,
                          size_t bytes_transferred)
    {
        if (!error)
        {
            std::cout.write(reply_, bytes_transferred);
            std::cout.flush();
            socket_.async_read_some(boost::asio::buffer(reply_, max_length),
                                    boost::bind(&WoprTerminal::listen_on_server, this,
                                                boost::asio::placeholders::error,
                                                boost::asio::placeholders::bytes_transferred));
        }
        else
        {
            /*
             * WOPR server on Unix-like systems:
             * ---------------------------------
             * Category == ssl && error == 335544539 => connection closed.
             * Category == misc && error == 2 => EOF => connection closed.
             *
             * WOPR server on Windows:
             * -----------------------
             * Category == system && error == 104 => connection reset by peer.
             * Category == system && error == 10054 => connection closed.
             */
            if ((error.category() == boost::asio::error::get_ssl_category() && error.value() == 335544539) ||
                    (error.category() == boost::asio::error::get_misc_category() && error.value() == 2) ||
                    (error.category() == boost::asio::error::get_system_category() && error.value() == 104) ||
                    (error.category() == boost::asio::error::get_system_category() && error.value() == 10054))
            {
                if (disconnect_)
                {
                    std::cout << "*** DISCONNECTED ***" << std::endl << std::endl;
                }
                else
                {
                    std::cout << "*** CONNECTION CLOSED BY REMOTE HOST ***" << std::endl << std::endl;
                }
                destroy_me();
            }
            else
            {
                std::cout << "Read failed: " << error.message() << std::endl;
                std::cout << error.category().name() << ':' << error.value() << std::endl;
                std::cout << "SSL: " << boost::asio::error::get_ssl_category().name() << std::endl;
                std::string err = error.message();
                if (error.category() == boost::asio::error::get_ssl_category())
                {
                    err = std::string(" (")
                            + boost::lexical_cast<std::string>(ERR_GET_LIB(error.value())) + ","
                            + boost::lexical_cast<std::string>(ERR_GET_FUNC(error.value())) + ","
                            + boost::lexical_cast<std::string>(ERR_GET_REASON(error.value())) + ") "
                            ;
                    char buf[128];
                    ::ERR_error_string_n(error.value(), buf, sizeof (buf));
                    err += buf;
                }
                std::cout << err << std::endl;
            }
        }
    }

    void handle_read_input(const boost::system::error_code& error,
                           std::size_t length)
    {
        if (!error)
        {
            // Read the entered command from input stream
            // into buffer request_.
            input_buffer_.sgetn(request_, length - 1);
            // The input buffer will be reused and async_read_until
            // will stop processing as soon as a \n is "detected"
            // in the input buffer.
            // Therefore we need to remove \n from input buffer.
            // Otherwise async_read_until will return immediately
            // with an empty command. A never ending loop would be
            // the result!
            input_buffer_.consume(1);
            // Ensure that our string in the request buffer is
            // terminated by a null byte.
            request_[length - 1] = '\0';

            // Check for know commands.
            std::string request(request_);
            boost::trim(request);
            boost::to_upper(request);
            if (request.compare("DISCONNECT") == 0)
            {
                disconnect_ = true;
            }

            // Send request to WOPR Server.
            boost::asio::async_write(socket_,
                                     boost::asio::buffer(request_, length),
                                     boost::bind(&WoprTerminal::handle_write, this,
                                                 boost::asio::placeholders::error));
        }
        else if (error == boost::asio::error::not_found)
        {
            // This should not happen!
            // max_length bytes read but no new line
            // => print an error message and drop read data
            std::cerr << "*** MAXIMUM NUMBER OF CHARACTERS READ ***"
                    << "***   BUT RETURN KEY WAS NOT PRESSED  ***"
                    << "***  DROPPING DATA AS UNKNOWN COMMAND ***"
                    << std::endl;
        }
        else
        {
            if (error.category() != boost::asio::error::get_misc_category() || error.value() != 2)
            {
                // Reason is not EOF so print error message:
                std::cerr << "Error occurred while reading data from stdin: "
                        << error.message() << std::endl;
                std::cerr << error.category().name() << ':' << error.value() << std::endl;
            }
            destroy_me();
        }
    }

    void handle_write(const boost::system::error_code& error)
    {
        if (!error)
        {
            if (!disconnect_)
            {
                // Read a line of input entered by the user.
                boost::asio::async_read_until(input_, input_buffer_, '\n',
                                              boost::bind(&WoprTerminal::handle_read_input, this,
                                                          boost::asio::placeholders::error,
                                                          boost::asio::placeholders::bytes_transferred));
            }
        }
        else
        {
            std::cerr << "Error occurred while transfering data to server: "
                    << error.message() << std::endl;
            std::cerr << error.category().name() << ':' << error.value() << std::endl;
            destroy_me();
        }
    }

    void destroy_me()
    {
        io_service_.stop();
    }

private:
    boost::asio::ssl::stream<boost::asio::ip::tcp::socket> socket_;
    boost::asio::posix::stream_descriptor input_;
    boost::asio::streambuf input_buffer_;
    char request_[max_length];
    char reply_[max_length];
    bool disconnect_;
    boost::asio::io_service& io_service_;
};

int main(int argc, char* argv[])
{
    boost::program_options::options_description desc(
                                                     "Usage: terminal [--host <host> | -h <host>] [--port <port> | -p <port>]\\\n"
                                                     "                [--cert <file> | -c <file>]\n"
                                                     "       terminal --help\n"
                                                     "       terminal --version\n"
                                                     "where"
                                                     );
    desc.add_options()
            ("help", "Print this help message.")
            ("version", "Print version information and exit.")
            ("host,h", boost::program_options::value<std::string>(), "Host to connect to. (Default: localhost)")
            ("port,p", boost::program_options::value<std::string>(), "Port to connect to. (Default: 7777)")
            ("cert,c", boost::program_options::value<std::string>(), "Certificate file for SSL. (Default: server.crt)")
            ;
    boost::program_options::variables_map vm;

    try
    {
        boost::program_options::store(boost::program_options::parse_command_line(argc, argv, desc), vm);
        boost::program_options::notify(vm);

        if (vm.count("help"))
        {
            std::cout << desc << std::endl;
            return 1;
        }

        if (vm.count("version"))
        {
            std::cout << "WOPR Terminal" << std::endl
                    << "Version " << VERSION << std::endl;
            return 0;
        }

        boost::asio::io_service io_service;
        std::string cert_file("server.crt");
        std::string host("localhost");
        std::string port("7777");

        if (vm.count("cert"))
        {
            cert_file = vm["cert"].as<std::string>();
        }

        if (vm.count("host"))
        {
            host = vm["host"].as<std::string>();
        }

        if (vm.count("port"))
        {
            port = vm["port"].as<std::string>();
        }

        boost::asio::ip::tcp::resolver resolver(io_service);
        boost::asio::ip::tcp::resolver::query query(host, port);
        boost::asio::ip::tcp::resolver::iterator iterator = resolver.resolve(query);

        boost::asio::ssl::context ctx(io_service, boost::asio::ssl::context::sslv23);
        ctx.set_verify_mode(boost::asio::ssl::context::verify_peer);
        ctx.load_verify_file(cert_file);

        WoprTerminal terminal(io_service, ctx, iterator);

        io_service.run();
    }
    catch (std::exception& e)
    {
        std::cerr << "Exception: " << e.what() << std::endl;
    }

    return 0;
}
