/*
 * Copyright (c) 2009-2015 Oliver Mueller [http://oliver-mueller.com]
 * 
 * This file is part of WOPR.
 * 
 * $Id: WoprSession.cpp 74 2015-10-01 14:50:55Z oliver $
 * 
 * WOPR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <algorithm>
#include <cstring>
#include <string>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <boost/algorithm/string.hpp>
#include <boost/date_time.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/bind.hpp>
#include "WoprSession.h"
#include "config.h"

namespace wopr
{

const char *WoprSession::identification = "\r\nWOPR " VERSION "\r\n";
const char *WoprSession::prompt[2] = {"\r\nLOGON: ", // logon prompt
    "\r\n" // command prompt
};
unsigned WoprSession::prompt_len[2] = {0, 0};
const char *WoprSession::access_denied = "\r\n** IDENTIFICATION NOT RECOGNIZED **\r\n        ** ACCESS DENIED **\r\n";
const char *WoprSession::already_running = "\r\n** GAME ROUTINE RUNNING **\r\n";
CmdNResp WoprSession::cmd_n_resp[] = {
    { "LIST GAMES", 0,
        "\r\nFALKENS' MAZE\r\n"
        "BLACK JACK\r\n"
        "GIN RUMMY\r\n"
        "HEARTS\r\n"
        "BRIDGE\r\n"
        "CHECKERS\r\n"
        "CHESS\r\n"
        "POKER\r\n"
        "FIGHTER COMBAT\r\n"
        "GUERRILLA ENGAGEMENT\r\n"
        "DESERT WARFARE\r\n"
        "AIR-TO-GROUND ACTIONS\r\n"
        "THEATREWIDE TACTICAL WARFARE\r\n"
        "THEATREWIDE BIOTOXIC AND CHEMICAL WARFARE\r\n\r\n"
        "GLOBAL THERMONUCLEAR WAR\r\n", 0,
        0x00},
    { "FALKENS' MAZE", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "BLACK JACK", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "GIN RUMMY", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "HEARTS", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "BRIDGE", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "CHECKERS", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "CHESS", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "POKER", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "FIGHTER COMBAT", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "GUERRILLA ENGAGEMENT", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "DESERT WARFARE", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "AIR-TO-GROUND ACTIONS", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "THEATREWIDE TACTICAL WARFARE", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "THEATREWIDE BIOTOXIC AND CHEMICAL WARFARE", 0,
        WoprSession::access_denied, 0,
        0x04},
    { "GLOBAL THERMONUCLEAR WAR", 0,
        WoprSession::already_running, 0,
        0x04},
    { "TIC-TAC-TOE", 0,
        "\r\nA STRANGE GAME.\r\nTHE ONLY WINNING MOVE IS\r\nNOT TO PLAY.\r\n", 0,
        0x04},
    { "HELP LOGON", 0,
        "\r\nHELP NOT AVAILABLE\r\n", 0,
        0x00},
    { "HELP GAMES", 0,
        "\r\n'GAMES' REFERS TO MODELS, SIMULATIONS AND GAMES\r\n"
        "WHICH HAVE TACTICAL AND STRATEGIC APPLICATIONS.\r\n", 0,
        0x00},
    { "LOGOFF", 0,
        "\r\nGOODBYE\r\nCLOSING USER SESSION\r\n", 0,
        0x06},
    { "DISCONNECT", 0,
        "\r\nGOODBYE\r\nCLOSING CONNECTION\r\n\r\n", 0,
        0x16},
    { "JOSHUA", 0,
        "\r\nGREETINGS PROFESSOR FALKEN.\r\n", 0,
        0x09},
    { "MCKITTRICK", 0,
        "\r\nGREETINGS DOCTOR MCKITTRICK.\r\n", 0,
        0x09},
    { "", 0, "", 0, 0x00}
};
const char *WoprSession::unknown_cmd =
        "\r\nUNKNOWN COMMAND\r\n\r\n";
unsigned WoprSession::unknown_cmd_len = 0;
const char *WoprSession::unknown_user =
        "\r\nIDENTIFICATION NOT RECOGNIZED BY SYSTEM\r\n"
        "--CONNECTION TERMINATED--\r\n\r\n";
unsigned WoprSession::unknown_user_len = 0;

WoprSession::WoprSession(boost::asio::io_service& io_service, boost::asio::ssl::context& context, unsigned client_number, bool join, bool multi_threading, unsigned delay)
: io_service_(io_service), socket_(io_service, context)
{
    destroying_ = false;
    join_ = join;
    multi_threading_ = multi_threading;
    delay_ = delay;
    logon_ = true;
    shutdown_ = false;
    std::ostringstream out;
    out << "["
            << boost::posix_time::to_iso_string(boost::posix_time::second_clock::local_time())
            << ":" << std::setfill('0')
            << std::setw(sizeof (unsigned) * 2)
            << std::hex
            << client_number << "] ";
    client_id_ = out.str();
}

WoprSession::~WoprSession()
{
}

void WoprSession::start()
{
    WOPR_LOG(kInfo) << client_id_ << "Session started.";
    socket_.async_handshake(boost::asio::ssl::stream_base::server,
                            boost::bind(&WoprSession::handle_handshake, this,
                                        boost::asio::placeholders::error));
}

void WoprSession::handle_handshake(const boost::system::error_code& error)
{
    if (!error)
    {
        WOPR_LOG(kInfo) << client_id_
                << "Handshake successful. Sending prompt to client.";

        std::string logon(prompt[0]);
        boost::asio::async_write(socket_,
                                 boost::asio::buffer(logon, logon.length()),
                                 boost::bind(&WoprSession::handle_write_logon, this,
                                             boost::asio::placeholders::error));
    }
    else
    {
        destroy_me();
    }
}

void WoprSession::handle_write_logon(const boost::system::error_code& error)
{
    if (!error)
    {
        WOPR_LOG(kInfo) << client_id_
                << "Handshake successful. Waiting for incoming messages.";
        socket_.async_read_some(boost::asio::buffer(data_, max_length),
                                boost::bind(&WoprSession::handle_read, this,
                                            boost::asio::placeholders::error,
                                            boost::asio::placeholders::bytes_transferred));
    }
    else
    {
        destroy_me();
    }

}

void WoprSession::handle_read(const boost::system::error_code& error,
                              size_t bytes_transferred)
{
    if (!error)
    {
        // assign data to std::string
        std::string d;
        d.assign(data_, strnlen(data_, bytes_transferred));

        if (multi_threading_) {
            // open thread for processing
            WOPR_LOG(kInfo) << client_id_
                    << "Starting thread for handling message \"" << d << "\"";
            threads_.create_thread(boost::bind(boost::bind(&WoprSession::process_data, this, d)));
        } else {
            // process data in this thread
            process_data(d);
        }

        // read new data
        WOPR_LOG(kInfo) << client_id_ << "Reading from socket.";
        socket_.async_read_some(boost::asio::buffer(data_, max_length),
                                boost::bind(&WoprSession::handle_read, this,
                                            boost::asio::placeholders::error,
                                            boost::asio::placeholders::bytes_transferred));
    }
    else
    {
        destroy_me();
    }
}

void WoprSession::handle_thread_done(const boost::system::error_code& error)
{
    if (!error)
    {
        WOPR_LOG(kInfo) << client_id_
                << "Data transfer to client was successful.";
    }
    else
    {
        WOPR_LOG(kError) << client_id_
                << "Error occurred while transfer of data to client: "
                << error.message();
        destroy_me();
    }
}

void WoprSession::process_data(std::string data)
{
    WOPR_LOG(kInfo) << client_id_ << "Processing message \"" << data << "\"";
    
    // Simulate long processing?
    if (delay_ != 0) {
        WOPR_LOG(kInfo) << client_id_ << "BEGIN of simulation of long processing.";
        boost::this_thread::sleep(boost::posix_time::milliseconds(delay_));
        WOPR_LOG(kInfo) << client_id_ << "END of simulation of long processing.";
    }
    // trim data
    boost::trim(data);
    collapse_spaces(data);
    boost::to_upper(data);

    // process received commands
    bool nocmd = true;
    int n;
    for (n = 0; cmd_n_resp[n].cmd[0] != '\0'; n++)
    {
        if (data.compare(cmd_n_resp[n].cmd) == 0)
        {
            if (logon_ && (cmd_n_resp[n].status & STA_LGNOLY))
            { // only at logged on state
                break;
            }
            if (!logon_ && (cmd_n_resp[n].status & STA_LGFOLY))
            { // only at logged off state
                break;
            }
            nocmd = false;
            if (cmd_n_resp[n].status & STA_LOGSET) // set logged on state?
                logon_ = false;
            if (cmd_n_resp[n].status & STA_LOGUNS) // set logged off state?
                logon_ = true;
            if (cmd_n_resp[n].status & STA_SHTDWN) // set shutdown?
                shutdown_ = true;
            break;
        }
    }

    // generate response string
    std::string response;
    if (nocmd)
    {
        if (logon_)
        {
            response = unknown_user;
            shutdown_ = true;
        }
        else
        {
            response = unknown_cmd;
        }
    }
    else
    {
        response = cmd_n_resp[n].resp;
    }

    if (!shutdown_)
    {
        response += prompt[(logon_ ? 0 : 1)];
    }

    WOPR_LOG(kInfo) << client_id_ << "Sending response \"" << response << "\"";
    boost::asio::async_write(socket_,
                             boost::asio::buffer(response, response.length()),
                             boost::bind(&WoprSession::handle_thread_done, this,
                                         boost::asio::placeholders::error)
                             );
    if (shutdown_)
    {
        join_ = true;
        // We close the socket here which will cause asyn_read or asyn_write
        // in the main thread to call destroy_me(). So we avoid having side
        // effects which might lead to segmentation violations.
        // Reason: We would "delete this;" in this worker thread. Afterwards
        // due to code like socket_.asyn_read_some(...) in the main thread
        // a segmentation fault occurres becasue socket_ is no longer
        // allocated.
        WOPR_LOG(kInfo) << client_id_ << "Disconnecting. Closing socket.";
        socket().close();
    }
}

void WoprSession::destroy_me()
{
    // If we are in single-threading mode we won't start any threads.
    // So if aren't staring threads we do not need to join them.
    if (!multi_threading_)
        return;
    
    // We are operating in a multi-threaded environment.
    // If one thread runs in the situation of a lost or closed connection
    // all others will, too.
    // Therefore we have to take care that this destroy_me() method will
    // do its work only once! Otherwise a segmentation fault will be the
    // result of the most-inner returning method call of destroy_me(),
    // because we delete the WoprSession object but have a call stack
    // operating on it.
    if (destroying_)
        return;
    destroying_ = true;
    if (join_)
    {
        WOPR_LOG(kInfo) << client_id_ << "Starting thread to join all threads.";
        // Start a thread which will join all threads and finally deletes
        // the session object. This is necessary because we are currently
        // in the thread which controls asynchronous operation.
        // In other words: We are in the thread where io_service.run() is
        // executed. If we wait here for all threads to join, we will block
        // the entire asynchronous processing.
        // Note: This new thread must not be a member of our thread group
        // threads_ because it will wait for them to join!
        boost::thread(boost::bind(&WoprSession::join_all, this));
    }
    else
    {
        WOPR_LOG(kInfo) << client_id_ << "Interrupting all threads.";
        // Interrupting of all threads can be done here. It should not
        // block our asynchronous work too long.
        // If you have a bad feeling about this: Start a separate thread
        // for interrupt_all() here, too.
        threads_.interrupt_all();
        if (socket().is_open())
        {

            WOPR_LOG(kInfo) << client_id_ << "Closing socket.";
            socket().close();
        }
        WOPR_LOG(kInfo) << client_id_ << "Destroying session.";
        delete this;
    }
}

void WoprSession::join_all()
{
    // Our asynchronous join all thread for a graceful finish of our
    // threads.
    WOPR_LOG(kInfo) << client_id_ << "Wait for all threads to join.";
    threads_.join_all();
    WOPR_LOG(kInfo) << client_id_ << "All threads joined.";
    if (socket().is_open())
    {
        WOPR_LOG(kInfo) << client_id_ << "Closing socket.";
        socket().close();
    }
    WOPR_LOG(kInfo) << client_id_ << "Destroying session.";
    delete this;
}

bool WoprSession::both_spaces(char left, char right)
{
    return (isspace(left) && isspace(right));
}

inline void WoprSession::collapse_spaces(std::string& str)
{
    // Substitute a sequence of spaces and/or tabs fore one space/tab:
    str.erase(std::unique(str.begin(), str.end(), WoprSession::both_spaces), str.end());
    // Replace all tabs with spaces:
    std::replace(str.begin(), str.end(), '\t', ' ');
}

} // namespace wopr
