/*
 * Copyright (c) 2009-2015 Oliver Mueller [http://oliver-mueller.com]
 * 
 * This file is part of WOPR.
 * 
 * $Id: WoprSession.h 70 2015-07-29 13:07:47Z oliver $
 * 
 * WOPR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef WOPRSESSION_H
#define	WOPRSESSION_H

#include <string>
#include <boost/system/error_code.hpp>
#include <boost/asio.hpp>
#include <boost/asio/ssl.hpp>
#include <boost/thread.hpp>
#include "WoprLogger.h"

namespace wopr {

typedef boost::asio::ssl::stream<boost::asio::ip::tcp::socket> ssl_socket;

#define STA_LOGSET    0x01
#define STA_LOGUNS    0x02
#define STA_LGNOLY    0x04
#define STA_LGFOLY    0x08
#define STA_SHTDWN    0x10

typedef struct
{
    const char *cmd;
    unsigned cmd_len;
    const char *resp;
    unsigned resp_len;
    short status;
} CmdNResp;


class WoprSession
{
public:
    WoprSession(boost::asio::io_service& io_service, boost::asio::ssl::context& context, unsigned client_number, bool join, bool multi_threading, unsigned delay);
    virtual ~WoprSession();

    ssl_socket::lowest_layer_type& socket();
    void start();

protected:
    void handle_handshake(const boost::system::error_code& error);
    void handle_write_logon(const boost::system::error_code& error);
    void handle_read(const boost::system::error_code& error, size_t bytes_transferred);
    void handle_thread_done(const boost::system::error_code& error);
    void process_data(std::string data);

    void destroy_me();
    void join_all();
    
    static bool both_spaces(char left, char right);
    void collapse_spaces(std::string &str);

private:
    boost::asio::io_service& io_service_;

    ssl_socket socket_;
    boost::thread_group threads_;
    bool join_;
    bool destroying_;

    enum
    {
        max_length = 1024
    };
    char data_[max_length];
    
    std::string client_id_;

    // String constants for WOPR responses:
    static const char *identification;
    static const char *prompt[2];
    static unsigned prompt_len[2];
    static const char *unknown_cmd;
    static unsigned unknown_cmd_len;
    static const char *unknown_user;
    static unsigned unknown_user_len;
    static CmdNResp cmd_n_resp[];
    static const char *access_denied;
    static const char *already_running;
    
    // Show logon prompt?
    bool logon_;
    // Shutdown planned?
    bool shutdown_;
    // Multi-threading or single-threading?
    bool multi_threading_;
    // Simulate long processing?
    unsigned delay_;
};

inline ssl_socket::lowest_layer_type& WoprSession::socket()
{
    return socket_.lowest_layer();
}

} // namespace wopr

#endif	/* WOPRSESSION_H */
