/*
 * Copyright (c) 2009-2015 Oliver Mueller [http://oliver-mueller.com]
 * 
 * This file is part of WOPR.
 * 
 * $Id: wopr.cpp 70 2015-07-29 13:07:47Z oliver $
 * 
 * WOPR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <boost/program_options.hpp>
#include <boost/asio.hpp>
#include <boost/asio/ssl.hpp>

#include "WoprServer.h"
#include "config.h"

int main(int argc, char* argv[])
{
    boost::program_options::options_description desc(
                                                     "Usage: wopr {--port <port> | -p <port>} \\\n"
                                                     "            [--key <file> | -k <file>] [--Pass <file> | -w <file>] \\\n"
                                                     "            [--cert <file> | -c <file>] [--join | -j] [--ipv6 | -6] \\\n"
                                                     "            [--delay <milliseconds>] [--nothreads]\n"
                                                     "       wopr --help\n"
                                                     "       wopr --version\n"
                                                     "where"
                                                     );

    desc.add_options()
            ("help", "Print this help message and quit.")
            ("version", "Print version number and quit.")
            ("port,p", boost::program_options::value<int>(), "Listen on given port for incoming SSL connections.")
            ("key,k", boost::program_options::value<std::string>(), "Key file for SSL. (Default: server.key)")
            ("cert,c", boost::program_options::value<std::string>(), "Chain of certificates file for SSL. (Default: server.crt)")
            ("Pass,P", boost::program_options::value<std::string>(), "Optional file containing the password for key file.")
            ("join,j", "Let all threads of a session join at finish instead of interrupting them.")
            ("ipv6,6", "Enable IPv6.")
            ("delay,d", boost::program_options::value<unsigned>(), "Simulate a long processing of requests by delaying processing by given milliseconds.")
            ("nothreads,n", "Disable multi-threading. Run all operations in just one thread.")
            ;

    boost::program_options::variables_map vm;

    try
    {
        boost::program_options::store(boost::program_options::parse_command_line(argc, argv, desc), vm);
        boost::program_options::notify(vm);

        if (vm.count("help"))
        {
            std::cout << desc << std::endl;
            return 1;
        }

        if (vm.count("version"))
        {
            std::cout << "WOPR version " << VERSION << std::endl;
            return 0;
        }

        // Process certificate options.
        boost::asio::io_service io_service;
        std::string key_file("server.key");
        std::string cert_file("server.crt");
        std::string password("");

        if (vm.count("key"))
        {
            key_file = vm["key"].as<std::string>();
        }
        if (vm.count("cert"))
        {
            cert_file = vm["cert"].as<std::string>();
        }
        if (vm.count("Pass"))
        {
            std::ifstream pass_file;
#if defined(__xlC__) || defined(__ibmxl__) || defined(__SUNPRO_CC) || defined(LEGACY_GCC)
            pass_file.open(vm["Pass"].as<std::string>().c_str());
#else
            pass_file.open(vm["Pass"].as<std::string>());
#endif
            char buffer[1024];
            pass_file.getline(buffer, 1024);
            password = std::string(buffer);
            pass_file.close();
        }
        bool join = (vm.count("join") ? true : false);

        // Set port listening on.
        if (!vm.count("port"))
        {
            std::cerr << "You have to specify at least a TCP port with --port!" << std::endl;
            return 1;
        }

        // Process IPv6 options.
        bool ipv6 = (vm.count("ipv6") ? true : false);

        // Simulate long processing?
        unsigned delay;
        if (!vm.count("delay"))
        {
            delay = 0;
        }
        else
        {
            delay = vm["delay"].as<unsigned>();
        }
        
        // Disable multi-threading?
        bool multi_threading = (!vm.count("nothreads"));

        // Starting server
        std::cout << "WOPR Server" << std::endl;
        std::cout << "Version " << VERSION << std::endl << std::endl;
        std::cout << "Listening on " << vm["port"].as<int>() << std::endl;
        std::cout << "Using key in " << key_file << std::endl;
        std::cout << "Using chain of certificates from " << cert_file << std::endl;
        if (vm.count("pass"))
        {
            std::cout << "Using password for key from " << vm["Pass"].as<std::string>() << std::endl;
        }
        std::cout << "Processing model: " << (multi_threading ? "mutli-threading" : "single-threading") << std::endl;
        if (delay != 0)
        {
            std::cout << "Simulation of long processing is enabled." << std::endl
                    << "Delaying processing of requests by " << delay
                    << " milliseconds." << std::endl;
        }
        std::cout << std::endl;
        wopr::WoprServer wopr_server(io_service, vm["port"].as<int>(), key_file, password, cert_file, join, ipv6, multi_threading, delay);
        io_service.run();
    }
    catch (std::exception& e)
    {
        std::cerr << "Exception: " << e.what() << "\n";
    }

    return 0;
}
